from typing import Union

import torch
from torch import nn

from .utils import get_appropriate_conv_layer


class Chomp(nn.Module):
    """
    cuts padding part of sequence
    inspired by https://github.com/locuslab/TCN
    """
    def __init__(self, chomp_size: int, dims: int=1):
        """
        args:
            `chomp_size` - padding size to cut off
            `dims` - number of working dimensionalities, which needed to be chomped
        """
        super().__init__()
        self.chomp_size = chomp_size
        if dims not in (1, 2, 3):
            raise NotImplementedError(f"Chomp layer for {dims = } not implemented")
        self.dims = dims

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.dims == 1:
            return x[..., : - self.chomp_size].contiguous()
        if self.dims == 2:
            return x[..., : - self.chomp_size, : - self.chomp_size].contiguous()
        if self.dims == 3:
            return x[..., : - self.chomp_size, : - self.chomp_size,  : - self.chomp_size].contiguous()

class TemporalBlock(nn.Module):
    """
    combination of (convolutional layer, chomp, relu, dropout) repeated `layers` times
    adds additional convolutional layer if needed to downsample number of channels
    inspired by https://github.com/locuslab/TCN
    """
    def __init__(self, n_inputs: int, n_outputs: int, kernel_size: Union[int, tuple[int]],
                  stride: Union[int, tuple[int]], dilation: Union[int, tuple[int]], padding: Union[int, tuple[int]],
                  dropout: int = 0.2, dims: int = 1, layers: int = 2):
        super().__init__()

        conv_layer = get_appropriate_conv_layer(dims)
        self.padding = padding
        self.dropout = dropout

        net = []
        for i in range(layers):
            net.append(torch.nn.utils.weight_norm(conv_layer(
                (n_inputs if i == 0 else n_outputs), n_outputs, kernel_size, stride=stride, padding=padding, dilation=dilation)))
            if padding > 0:
                net.append(Chomp(padding, dims))
            net.append(nn.ReLU())
            if dropout > 0:
                net.append(nn.Dropout(dropout))
        self.net = nn.ModuleList(net)

        self.downsample = conv_layer(n_inputs, n_outputs, 1) if n_inputs != n_outputs else None
        self.relu = nn.ReLU()
        self.init_weights()

    def init_weights(self):
        """
        sets normal weight distribution for convolutional layers
        """
        for i in range(0, len(self.net), 2 + (self.dropout > 0) + (self.padding > 0)):
            self.net[i].weight.data.normal_(0, 0.5)

        if self.downsample is not None:
            self.downsample.weight.data.normal_(0, 0.5)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        `x` in format [batch_size, channels, *other_dims]
        """
        out = x
        for i in range(len(self.net)):
            out = self.net[i](out)

        res = x if self.downsample is None else self.downsample(x)
        return out, self.relu(out + res)



import pytest

import numpy as np

import torch
from torch import nn

from timediffusion import count_params, kl_div, DimUniversalStandardScaler



@pytest.mark.parametrize(
    "arr",
    [
        np.sin(np.arange(10)) * 10,
        torch.arange(10).float(),
        np.array([np.nan, 1.0, 2.0]),
        torch.Tensor([torch.nan, 1.0, 2.0])
    ]
)
def test_duscaler(arr):
    if isinstance(arr, np.ndarray):
        nan_mask = np.isnan(arr)
    else:
        nan_mask = torch.isnan(arr)
    mask = ~ nan_mask

    scaler = DimUniversalStandardScaler()
    tarr = scaler.fit_transform(arr)
    tarr1 = scaler.transform(arr)
    rarr = scaler.inverse_transform(tarr)
    assert abs((tarr[mask] - tarr1[mask]).mean()) < scaler.eps
    assert abs((rarr[mask] - arr[mask]).mean()) < scaler.eps

@pytest.mark.parametrize(
    "x,y",
    [
        (np.sin(np.arange(10)), np.arange(10)),
        (torch.sin(torch.arange(10)), torch.arange(10)),
        pytest.param(np.arange(10), torch.arange(10), marks=pytest.mark.xfail)
    ],
)
def test_kl_div(x, y):
    kl_div(x, y)

@pytest.mark.parametrize(
    "in_features,out_features",
    [(100, 100), (200, 200)],
)
def test_count_params_linear(in_features, out_features):
    linear = nn.Linear(in_features, out_features)
    assert count_params(linear) == in_features * out_features + out_features

@pytest.mark.parametrize(
    "in_channels,out_channels,kernel_size,groups",
    [
        (3, 24, 3, 1),
        (1, 10, 2, 1),
        (3, 12, 4, 3)
     ],
)
def test_count_params_conv(in_channels, out_channels, kernel_size, groups):
    conv = nn.Conv1d(in_channels=in_channels, out_channels=out_channels,
         kernel_size=kernel_size, groups=groups)
    assert count_params(conv) == out_channels * (in_channels // groups) * kernel_size + out_channels



from typing import Union

import numpy as np

import torch
from torch import nn


def count_params(model: nn.Module) -> int:
    """
    counts number of model parameters
    """
    res = 0
    for param in model.parameters():
        res += np.prod(param.shape)
    return res


def get_appropriate_conv_layer(dims: int) -> nn.Module:
    """
    returns appropriate convolutional layer for certain number of dimensionalities
    """
    if dims not in (1, 2, 3):
        raise NotImplementedError(f"Convolutional layer for dimensionalty {dims} not implemented")
    return {1: nn.Conv1d, 2: nn.Conv2d, 3: nn.Conv3d}[dims]


def kl_div(x: Union[np.ndarray, torch.Tensor], y: Union[np.ndarray, torch.Tensor],
           eps: float = 1e-6):
    """
    Calculates kl_div on min-max version of x, y
    
    args:
        `x` - input array
        `y` - reference array
    """
    if type(x) is not type(y):
        raise ValueError(f"input arrays for kl_div should be same type, got {type(x)} and {type(y)}")

    if type(x) is np.ndarray:
        clip = lambda arr: np.clip(arr, a_min=eps, a_max=1)
        log = np.log
    elif type(x) is torch.Tensor:
        clip = lambda arr: torch.clip(arr, min=eps, max=1)
        log = torch.log
    else:
        raise NotImplementedError(f"kl_div array type should be numpy.array or torch.tensor`, got {type(x)}")
    
    # biased min-max to get rid off zeros
    x = (x - x.min())  / clip(x.max() - x.min())
    y = (y - y.min()) / clip(y.max() - y.min())
    x = clip(x)
    y = clip(y)

    # to probs
    x /= x.sum()
    y /= y.sum()

    return (log(x / y) * x)


class DimUniversalStandardScaler:
    """
    Universal class for normal scaling data
    """
    def __init__(self, eps=1e-9):
        self.eps = eps

    def fit(self, data):
        if isinstance(data, torch.Tensor):
            mask = ~ torch.isnan(data)
            self.mu = data[mask].mean().item()
            self.std = data[mask].std().item()
        else:
            mask = ~ np.isnan(data)
            self.mu = data[mask].mean()
            self.std = data[mask].std()

    def transform(self, data):
        return (data - self.mu) / (self.std + self.eps)

    def fit_transform(self, data):
        self.fit(data)
        return self.transform(data)

    def inverse_transform(self, data):
        return data * (self.std + self.eps) + self.mu



from typing import Union
from abc import ABC, abstractmethod

import numpy as np

import torch
from torch import nn
import torch.nn.functional as F

from .utils import get_appropriate_conv_layer
from .layers import TemporalBlock, Chomp


class TimeDiffusionModel(nn.Module, ABC):
    """
    Abstract class for all base models
    """
    @abstractmethod
    def __init__(self, *args, **kwargs):
        super().__init__()

    @abstractmethod
    def forward(self, *args, **kwards):
        pass


class TimeDiffusionProjector(TimeDiffusionModel):
    """
    convolutional network, used as projector in TD
    consists of temporal blocks with exponentially increasing padding/dilation parameters
    """
    def __init__(self, input_dims: Union[list[int], tuple[int]], max_deg_constraint: Union[int, None] = None,
                  kernel_size: int = 2, conv_filters: int = 128, base_dropout: float = 0.05):
        """
        args:

            `input_dims` - [channels, *dims]
                needed for dynamical network building
                best way to pass it as `x.shape` (without batches)

            `max_deg_constraint` - constraint to lessen network size, if not big enough will worsen model quality
                number of temporal blocks in network will be (1 + max_deg_constraint) maximum
                if None: ignores parameter

            `kernel_size` - base kernel size for dilated convolutions

            `conv_filters` - number of convolutional filters for each layer

            `base_dropout` - dropout for first temporal block
        """
        super().__init__()

        self.kernel_size = kernel_size
        self.input_dims = input_dims
        self.dims = len(input_dims) - 1
        self.channels = input_dims[0]
        self.max_seq = max(input_dims[1:])
        self.max_deg = int(np.ceil(np.log2(self.max_seq)))
        if max_deg_constraint is not None and max_deg_constraint < self.max_deg:
            print(f"For better TimeDiffusion performance it's recommended to use max_deg_constraint ", end="")
            print(f"with value{self.max_deg} for input with shape {input_dims}")
            self.max_deg = max_deg_constraint
            print(f"Setting current {self.max_deg = }")

        self.tcn = nn.ModuleList(
            [TemporalBlock(self.channels, conv_filters, 
                           kernel_size=1, stride=1, dilation=1, padding=0, dropout=base_dropout, dims=self.dims),
             *[TemporalBlock(conv_filters, conv_filters, 
                             kernel_size=kernel_size, stride=1, dilation=i, padding=i * (kernel_size - 1),
                               dropout=0.0, dims=self.dims)
                                        for i in [kernel_size ** i for i in range(self.max_deg + 1)]]
                                    ])
        
        self.last = get_appropriate_conv_layer(self.dims)(conv_filters, self.channels, kernel_size=1, stride=1, dilation=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        skip_acc = None
        for layer in self.tcn:
            skip, x = layer(x)
            if skip_acc is None:
                skip_acc = skip
            else:
                skip_acc += skip
        x = self.last(x + skip_acc)
        return x


class TimeDiffusionAttention(TimeDiffusionModel):
    """
    attention model, uses projectors to create (q, k, v) for vanilla attention layer
    """
    def __init__(self, *args, **params):
        """
        `args`, `params` - parameters for projectors
        """
        super().__init__()
        self.key_proj = TimeDiffusionProjector(*args, **params)
        self.val_proj = TimeDiffusionProjector(*args, **params)
        self.query_proj = TimeDiffusionProjector(*args, **params)

        self.input_dims = self.key_proj.input_dims
        self.dims = self.key_proj.dims

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # projections
        key = self.key_proj(x)
        val = self.val_proj(x)
        query = self.query_proj(x)

        is_batched = self.dims + 2 == len(key.size())
        mat_mul = torch.bmm if is_batched else torch.matmul

        # flattening last dimensionalities in case of 2D and 3D input
        # TODO: think of better solution
        if self.dims > 1:
            orig_shape = key.shape
            new_shape = list(key.size()[: - self.dims]) + [np.prod(key.size()[ - self.dims:])]

            key = key.view(new_shape)
            val = val.view(new_shape)
            query = query.view(new_shape)

        # vanilla attenion
        scores = mat_mul(query, key.transpose(- 2, - 1))
        weights = torch.nn.functional.softmax(scores, dim=1)
        attention = mat_mul(weights, val)

        # back to original shape in case of 2D and 3D input
        if self.dims > 1:
            attention = attention.view(orig_shape)

        return attention


class TimeDiffusionLiquid(TimeDiffusionModel):
    def __init__(self, input_dims: Union[list[int], tuple[int]], max_deg_constraint: Union[int, None] = None,
                  kernel_size: int = 2, conv_filters: int = 128, base_dropout: float = 0.05):
        """
        args:

            `input_dims` - [channels, *dims]
                needed for dynamical network building
                best way to pass it as `x.shape` (without batches)

            `max_deg_constraint` - constraint to lessen network size, if not big enough will worsen model quality
                number of temporal blocks in network will be (1 + max_deg_constraint) maximum
                if None: ignores parameter

            `kernel_size` - base kernel size for dilated convolutions

            `conv_filters` - number of convolutional filters for each layer

            `base_dropout` - dropout for first temporal block
        """
        super().__init__()
        
        self.kernel_size = kernel_size
        self.input_dims = input_dims
        self.dims = len(input_dims) - 1
        self.in_channels = input_dims[0]
        self.max_seq = max(input_dims[1:])
        self.max_deg = int(np.ceil(np.log2(self.max_seq)))
        if max_deg_constraint is not None and max_deg_constraint < self.max_deg:
            print(f"For better TimeDiffusion performance it's recommended to use max_deg_constraint ", end="")
            print(f"with value{self.max_deg} for input with shape {input_dims}")
            self.max_deg = max_deg_constraint
            print(f"Setting current {self.max_deg = }")

        conv_init = get_appropriate_conv_layer(self.dims)
        self.projector = conv_init(self.in_channels, conv_filters, kernel_size=1)
        self.base_dropout = nn.Dropout(base_dropout)

        self.conv = conv_init(conv_filters, conv_filters, kernel_size)
        self.conv_func = [F.conv1d, F.conv2d, F.conv3d][self.dims - 1]
        self.chomps = [Chomp(self.kernel_size ** i * (self.kernel_size - 1), self.dims) for i in range(self.max_deg + 1)]

        self.out_projector = conv_init(conv_filters, self.in_channels, kernel_size=1)

        self.__init_weights()

    def __init_weights(self):
        nn.init.kaiming_uniform_(self.projector.weight, a=0.01)
        nn.init.kaiming_uniform_(self.conv.weight, a=0.01)
        nn.init.kaiming_uniform_(self.out_projector.weight, a=0.01)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.projector(x)
        x = self.base_dropout(x)

        # as network uses weights of only one convolutional layer
        # we can reuse it for more dilation/padding combination on the fly
        cur_deg = int(np.ceil(np.log2(max(x.shape[- self.dims:]))))
        for i in range(len(self.chomps), cur_deg + 1):
            self.chomps.append(Chomp(self.kernel_size ** i * (self.kernel_size - 1), self.dims))

        # as network is much smaller, it's implemented in more functional way
        for i in range(cur_deg + 1):
            dilation = self.kernel_size ** i
            x = self.conv_func(x, self.conv.weight, self.conv.bias,
                                     padding=dilation * (self.kernel_size - 1), dilation=dilation)
            x = F.relu(x)
            x = self.chomps[i](x)

        x = self.out_projector(x)
        return x



from .utils import count_params, DimUniversalStandardScaler, kl_div
from .models import TimeDiffusionModel, TimeDiffusionProjector, TimeDiffusionAttention,\
                    TimeDiffusionLiquid
from .frameworks import TD

__all__ = [
    # useful functions
    "count_params",
    "kl_div",
    # data processing
    "DimUniversalStandardScaler",
    # abstract model class
    "TimeDiffusionModel",
    # models
    "TimeDiffusionProjector",
    "TimeDiffusionAttention",
    "TimeDiffusionLiquid",
    # frameworks
    "TD"
]



from typing import Union

from tqdm import tqdm

import numpy as np

import torch
from torch import nn

from .utils import count_params, DimUniversalStandardScaler, kl_div as _kl_div
from .models import TimeDiffusionProjector, TimeDiffusionAttention, TimeDiffusionModel
    

class TD(nn.Module):
    """
    Class provides a convenient framework for effectively working with TimeDiffusionProjector, encompassing all essential functions.
    """
    def __init__(self, base_model_init: TimeDiffusionModel = TimeDiffusionProjector,
                     verbose: bool = False, seed=42, *args, **params):
        """
        args (mostly same as TimeDiffusionProjector):

            `base_model_init` - base model init func, should inherited from TimeDiffusionModel

            `verbose` - whether to report number of model parameters

            `seed` - random seed for model parameters initialization

            `input_dims` - [channels, *dims]
                needed for dynamical network building
                best way to pass it as `x.shape` (without batches)

            `max_deg_constraint` - constraint to lessen network size, if not big enough will worsen model quality
                number of temporal blocks in network will be (1 + max_deg_constraint) maximum

            `conv_filters` - number of convolutional filters for each layer

            `base_dropout` - dropout for first temporal block
        """
        super().__init__()
        torch.random.manual_seed(seed)
        self.model = base_model_init(*args, **params)
        self.input_dims = self.model.input_dims
        self.is_fitted = False
        if verbose:
            print(f"Created model with {count_params(self):.1e} parameters")

    def dtype(self):
        return next(self.model.parameters()).dtype

    def device(self):
        return  next(self.model.parameters()).device

    def fit(self, example: Union[np.ndarray, torch.Tensor], mask: Union[None, np.ndarray, torch.Tensor] = None,
            epochs: int = 20, batch_size: int = 2, steps_per_epoch: int = 32,
            early_stopping_epochs: Union[None, int] = None,
            lr: float = 4e-4, distance_loss: Union[str, nn.Module] = "MAE",
            distribution_loss: Union[str, nn.Module] = "kl_div", distrib_loss_coef = 1e-2,
            verbose: bool = False, seed=42) -> list[float]:
        """
        simulates diffusion process for model training

        args:

            `example` - [sequence | image | video] in format (channels, *dims)

            `mask` - None for full model fitting on `example`
                or same shape as `example` for not fitting in points, that masked with 1

            `epochs` - number of training epochs

            `batch_size` - number of random noises to train on
                balance between (epochs - batch_size) means balance between (time - memory)
                more batch_size usually gives better results, but increasing epochs gives more significant improvement

            `steps_per_epoch` - number of diffusion steps to train each epoch

            `early_stopping_epochs` - whether to validate model after each epoch
                and stop model after `early_stopping_epochs` restoring quality decrease without improvement

            `lr` - learning rate

            `distance_loss` - main loss for fitting into input exampls
                should be either "MAE", "MSE"
                or pytorch nn.Module, that produces some kind of distance loss without dimensionality reduction

            `distribution_loss` - additional loss
                should be either `kl_div` for using built-in Kullback–Leibler divergence
                or pytorch nn.Module, that produces some kind of distributional loss without dimensionality reduction
                if pytorch module, it should take tensors with shape [batch_size, channels, *dims]

            `distrib_loss_coef` - scale of distribution loss in total loss

            `verbose` - whether to output training progress or not

            `seed` - random seed for fit method reproducibility

        returns:
            list of training losses (per step for each epoch)
        """
        # distance loss definition
        if isinstance(distance_loss, str):
            if distance_loss not in ("MAE", "MSE"): 
                raise NotImplementedError(f"Distance loss {distance_loss} doesn't exist")
            _mae = lambda x, y: (x - y).abs()
            _mse = lambda x, y: ((x - y) ** 2)
            distance_loss = {"MAE": _mae, "MSE": _mse}[distance_loss]
        elif not isinstance(distance_loss, nn.Module):
            raise NotImplementedError(f"Distance loss should be 'MAE', 'MSE' or nn.Module, got {type(distance_loss)}")

        # distribution loss definition
        if isinstance(distribution_loss, str):
            if distribution_loss != "kl_div":
                raise NotImplementedError(f"Distribution loss {distribution_loss} doesn't exist")
            distribution_loss = _kl_div
        elif not isinstance(distribution_loss, nn.Module):
            raise NotImplementedError(f"Distribution loss should be 'kl_div' or nn.Module got {type(distribution_loss)}")
        
        # mask check
        if mask is not None and mask.shape != example.shape:
            raise ValueError(f"Mask should None or the same shape as example, got {example.shape = } and {mask.shape = }")

        # scaling
        self.scaler = DimUniversalStandardScaler()
        train_tensor = torch.tensor(example, dtype=self.dtype(), device=self.device()).unsqueeze(0)
        train_tensor = self.scaler.fit_transform(train_tensor)
        X = train_tensor.repeat(batch_size, *[1] * (len(train_tensor.shape) - 1))

        if mask is not None:
            mask_tensor = ~ torch.tensor(mask, dtype=torch.bool, device=self.device()).unsqueeze(0)
            mask_tensor = mask_tensor.repeat(batch_size, *[1] * (len(mask_tensor.shape) - 1))

        optim = torch.optim.Adam(self.parameters(), lr=lr)
        losses = []
        if early_stopping_epochs is not None:
            val_losses = []
            val_noise = torch.rand(*X.shape, device=self.device(), dtype=self.dtype())

        torch.random.manual_seed(seed)
        for epoch in (tqdm(range(1, epochs + 1)) if verbose else range(1, epochs + 1)):
            self.model.train()

            noise = torch.rand(*X.shape, device=self.device(), dtype=self.dtype())
            # noise_level = torch.rand(X.shape).to(device=self.device(), dtype=self.dtype())
            # noise *= noise_level
            # scaling random noise with noise level gives additional training diversity and stability in some cases
            # TODO: further research in this area

            for step in range(steps_per_epoch):
                optim.zero_grad()
                y_hat = self.model(noise)
                # noise - y_hat -> X
                loss = distance_loss(noise - y_hat, X) + distrib_loss_coef * distribution_loss(y_hat, noise)
                loss = loss.mean() if mask is None else loss[mask_tensor].mean()
                loss.backward()
                optim.step()

                with torch.no_grad():
                    noise -= y_hat
                losses.append(loss.item())

            # validation
            if early_stopping_epochs is not None:
                self.model.eval()
                with torch.no_grad():
                    cur = val_noise.clone()
                    for step in range(steps_per_epoch):
                        cur -= self.model(cur)
                    val_losses.append(distance_loss(cur, X).mean().item())
                
                best_val_epoch = np.argmin(val_losses)
                if epoch - best_val_epoch - 1 >= early_stopping_epochs:
                    if verbose:
                        print(f"Due to early stopping fitting stops after {epoch}")
                        print(f"Val quality of {best_val_epoch} epoch {val_losses[best_val_epoch]: .1e}")
                        print(f"\tof current {val_losses[- 1]: .1e}")
                    break

        # saving some training parameters, could be useful in inference
        self.training_steps_per_epoch = steps_per_epoch
        self.training_example = example
        self.distance_loss = distance_loss
        self.distribution_loss = distribution_loss
        self.is_fitted = True

        return losses
    
    @torch.no_grad()
    def restore(self, example: Union[None, np.ndarray, torch.Tensor] = None, shape: Union[None, list[int], tuple[int]] = None,
                    mask: Union[None, np.ndarray, torch.Tensor] = None, steps: Union[None, int] = None,
                    seed: int = 42, verbose: bool = False) -> torch.Tensor:
        """
        recreates data using fitted model

        either `example` or `shape` should be provided

        3 possible workflows

            1 case of `shape`: model starts with random noise

            2 case of `example`: ignores `shape` and model starts with `example`

            3 case of `example` and `mask`: same as 2 case, but masked values persistent through diffusion process

        args:

            `example` - None or in format [channels, *dims], channels should be the same as in training example

            `shape` - None or in format [channels, *dims], channels should be the same as in training example

            `mask` - None or in format of `example`, zeros in positions, that needed to be persistent

            `steps` - steps for diffusion process, if None uses same value as in fit method

            `seed` - random seed, only necessary in case of providing only `shape`

            `verbose` - whether to output progress of diffusion process or not

        returns:
            result of diffusion process (torch.Tensor)
        """
        if not self.is_fitted:
            raise RuntimeError("Model isn't fitted")

        if example is None:
            if shape is None:
                raise ValueError("Either `example` or `shape` should be passed")

            torch.random.manual_seed(seed)
            X = torch.rand(*shape).to(device=self.device(), dtype=self.dtype())

            # no real meaning behind masking random noise
            # maybe for fun, but setting it here as None for stability
            mask = None
        else:
            if len(self.input_dims) != len(example.shape):
                raise ValueError(f"Model fitted with {len(self.input_dims)} dims, but got {len(example.shape)}")

            if self.input_dims[0] != example.shape[0]:
                raise ValueError(f"Model fitted with {self.input_dims[0]} channels, but got {example.shape[0]}")

            X = torch.tensor(example, device=self.device(), dtype=self.dtype())
            X = self.scaler.transform(X)

            if mask is not None:
                if mask.shape != example.shape:
                    raise ValueError(f"Mask should be same shape as example, got {example.shape = } {mask.shape = }")
                
                mask = torch.tensor(mask, device=self.device(), dtype=torch.bool)

            # provided example could have nan values
            nan_mask = torch.isnan(X)
            X[nan_mask] = torch.randn(nan_mask.sum(), device=X.device, dtype=X.dtype)

        steps = self.training_steps_per_epoch if steps is None else steps
        self.model.eval()
        for step in (tqdm(range(steps)) if verbose else range(steps)):
            preds = self.model(X)
            if mask is None:
                X -= preds
            else:
                X[mask] -= preds[mask]

        X = self.scaler.inverse_transform(X)
        return X

    def forecast(self, horizon: Union[int, tuple[int], list[int]], steps: Union[None, int] = None,
                    seed: int = 42, verbose: bool = False) -> torch.Tensor:
        """
        convinient version of `restore` to only get prediction for forecasting `horizon`
        uses trained sequence as masked (persistent) reference to forecast next values

        WORKS ONLY FOR 1D DATA

        args:

            `horizon` - forecasting horizon (e.g. number of values to forecast)

            `steps` - steps for diffusion process, if None uses same value as in fit method

            `seed` - random seed for reproducability

            `verbose` - whether to output progress of diffusion process or not

        returns:
            forecasted values (torch.Tensor in shape [channels, horizon])
        """
        if not self.is_fitted:
            raise RuntimeError("Model isn't fitted")
        if len(self.input_dims) != 2:
            raise NotImplementedError("forecast method works only for 1D data")
        
        np.random.seed(seed)
        example = np.append(self.training_example, np.zeros((self.input_dims[0], horizon)), axis=1)
        mask = np.zeros_like(example)
        mask[:, - horizon:] = 1

        res = self.restore(example=example, mask=mask, steps=steps, seed=seed, verbose=verbose)
        return res[:, - horizon:]

    @torch.no_grad()
    def synth(self, start=None, proximity: float = 0.9, samples: int = 8, batch_size: int = 8, 
            step_granulation : int = 10,
            seed: int = 42, verbose: bool = False) -> torch.Tensor:
        """
        generates synthetic data according to the proximity to the original example
            could start with random noise and denoise it to certain degree
            or start with provided samples and work with them

        args:

            `start` - diffusion process starts with 
                random noise in case None
                with provided samples [samples, channesl, *dims] in other cases

            `proximity` - similarity of syntetic samples to the fitted example
                should be in [0.0, 1.0] where 0.0 - random noise, 1.0 - full restored

            `samples` - number of synthetic samples to generate

            `batch_size` - number of sequences to generate at one diffusion process
                sets tradeoff between speed and memory of generation

            `step_granulation` - number of substeps to split usual step
                significantly increases computation time 
                    in tradeoff of better closeness of synthetic samples to selected proximity

            `seed` - random seed for reproducability

            `verbose` - whether to output progress or not

        returns:
            torch.Tensor of shape [samples, channels, *dims]
        """
        if not self.is_fitted:
            raise RuntimeError("Model isn't fitted")
        if not 0. <= proximity <= 1.:
            raise ValueError(f"proximity should be in [0, 1], got {proximity = }")
        dist = _kl_div
        gran_coef = 1 / step_granulation
        torch.random.manual_seed(seed)
        
        # proximity estimation of fitted model
        # TODO: think of better proximity mechanism
        if start is None:
            x = torch.rand(*self.input_dims, device=self.device(), dtype=self.dtype())
        else:
            x = torch.tensor(start[0], device=self.device(), dtype=self.dtype())
        ref = self.training_example.numpy() if type(self.training_example) is torch.Tensor else self.training_example
        scores = [dist(x.cpu().numpy(), ref).mean()]
        if verbose:
            print("Estimating fitted proximity...")

        self.model.eval()
        _range = range(self.training_steps_per_epoch * step_granulation)
        for _ in (tqdm(_range) if verbose else _range):
            preds = self.model(x) * gran_coef
            x -= preds
            scores.append(dist(x.cpu().numpy(), ref).mean())
        
        scores = 1 - np.array(scores)
        scores = (scores - scores.min()) / (scores.max() - scores.min())
        best_step = np.argmin(np.abs(scores - proximity))
        if verbose:
            print(f"Best granulated step is {best_step}")

        # generation
        res = []
        if verbose:
            print("Generating...")
        _range = range(0, samples, batch_size)
        for i in (tqdm(_range) if verbose else _range):
            if start is None:
                x = torch.rand(min(batch_size, samples - i),
                    *self.input_dims, device=self.device(), dtype=self.dtype())
            else:
                x = torch.tensor(start[i: i + batch_size], device=self.device(), dtype=self.dtype())
            for _ in range(best_step - 1):
                x -= self.model(x) * gran_coef
            res.append(x)

        res = torch.concat(res, dim=0)
        res = self.scaler.inverse_transform(res)

        return res



import pytest

import numpy as np

import torch
from torch import nn

from timediffusion import TD


@pytest.mark.parametrize("dims", [[1, 35], [1, 7, 7], [1, 5, 5, 5], [2, 35], [2, 7, 7], [2, 5, 5, 5]])
@pytest.mark.parametrize("mask_dropout", [None, 0.2])
class TestTD:
    def test_fit(self, dims, mask_dropout):
        model = TD(input_dims=dims)

        data = np.ones(dims)
        if mask_dropout is None:
            mask = None
        else:
            np.random.seed(42)
            mask = np.random.uniform(low=0., high=1.0, size=data.shape) < mask_dropout
        
        try:
            model.fit(data, mask=mask, early_stopping_epochs=5, epochs=1, batch_size=1, steps_per_epoch=2)
        except Exception as e:
            pytest.fail(f"TD.fit with {dims = } failed with exception: {e}")

    def test_restore(self, dims, mask_dropout):
        model = TD(input_dims=dims)

        data = np.ones(dims)
        if mask_dropout is None:
            mask = None
        else:
            np.random.seed(42)
            mask = np.random.uniform(low=0., high=1.0, size=data.shape) < mask_dropout
            data[mask] = np.nan

        try:
            model.fit(example=data, mask=mask, epochs=1, batch_size=1, steps_per_epoch=2)
            res = model.restore(data, mask=mask, steps=2)
        except Exception as e:
            pytest.fail(f"TD.restore with {dims = } failed with exception: {e}")

        assert (~ torch.isnan(res)).all()
        if mask is not None:
            # restoring with mask shouldn't change unmasked tokens
            assert np.allclose(res.numpy()[~ mask], data[~ mask])

    def test_forecast(self, dims, mask_dropout):
        if len(dims) > 2:
            return
        horizon = 3

        model = TD(input_dims=dims)

        data = np.ones(dims)
        if mask_dropout is None:
            mask = None
        else:
            np.random.seed(42)
            mask = np.random.uniform(low=0., high=1.0, size=data.shape) < mask_dropout

        try:
            model.fit(example=data, mask=mask, epochs=1, batch_size=1, steps_per_epoch=2)
            res = model.forecast(horizon, steps=2)
        except Exception as e:
            pytest.fail(f"TD.forecast with {dims = } failed with exception: {e}")

        assert len(res.shape) == 2
        assert res.shape[0] == dims[0]
        assert res.shape[1] == horizon
        assert (~ torch.isnan(res)).all()

    def test_synth(self, dims, mask_dropout):
        # using mask_dropout as a marker to start from noise or not
        start = mask_dropout

        model = TD(input_dims=dims)

        data = np.ones(dims)
        if start is not None:
            start = np.repeat(np.expand_dims(data, 0), 2, axis=0)
        samples = 2
        try:
            model.fit(example=data, epochs=1, batch_size=1, steps_per_epoch=2)
            if start is None:
                res = model.synth(proximity=0.7, step_granulation=2, samples=samples, batch_size=1)
            else:
                res = model.synth(start=start, proximity=0.7, step_granulation=2, samples=samples, batch_size=1)
        except Exception as e:
            pytest.fail(f"TD.synth with {dims = } and {start is None } failed with exception: {e}")
        
        assert len(res) == samples
        assert list(res[0].shape) == dims
        assert (~ torch.isnan(res)).all()



import pytest

import numpy as np

import torch
from torch import nn

from timediffusion import TimeDiffusionModel, TimeDiffusionProjector,\
                         TimeDiffusionAttention, TimeDiffusionLiquid


@pytest.mark.parametrize("model_init", [TimeDiffusionProjector, TimeDiffusionAttention, TimeDiffusionLiquid])
@pytest.mark.parametrize("dims", [[1, 35], [1, 7, 7], [1, 5, 5, 5], [2, 35], [2, 7, 7], [2, 5, 5, 5]])
@pytest.mark.parametrize("kernel_size", [2, 3])
class TestTimeDiffusionModel:
    def test_forward_pass(self, model_init, dims, kernel_size):
        model = model_init(input_dims=dims, kernel_size=kernel_size)
        assert isinstance(model, TimeDiffusionModel)

        # unbatched forward pass
        data = torch.ones(*dims)
        try:
            res = model(data)
        except Exception as e:
            pytest.fail(f"Unbatched forward pass of {type(model).__name__} with {dims = } failed with exception: {e}")

        assert data.shape == res.shape
        assert (~ torch.isnan(res)).all()
        
        # batched forward pass
        data = torch.ones(1, *dims)
        try:
            res = model(data)
        except Exception as e:
            pytest.fail(f"Batched forward pass of {type(model).__name__} with {dims = } failed with exception: {e}")

        assert data.shape == res.shape
        assert (~ torch.isnan(res)).all()

    def test_backward_pass(self, model_init, dims, kernel_size):
        model = model_init(input_dims=dims, kernel_size=kernel_size)
        assert isinstance(model, TimeDiffusionModel)

        # unbatched backward pass
        data = torch.ones(*dims)
        try:
            res = model(data)
            loss = (res - 1).mean()
            loss.backward()
        except Exception as e:
            pytest.fail(f"Unbatched backward pass of {type(model).__name__} with {dims = } failed with exception: {e}")

        assert (~ torch.isnan(loss)).all()
        
        # batched backward pass
        data = torch.ones(1, *dims)
        try:
            res = model(data)
            loss = (res - 1).mean()
            loss.backward()
        except Exception as e:
            pytest.fail(f"Batched backward pass of {type(model).__name__} with {dims = } failed with exception: {e}")

        assert (~ torch.isnan(loss)).all()
